"""Tasks related to interacting with the Tamr auxiliary service DF-connect"""
from dataclasses import dataclass
import json
import logging
import requests
from enum import Enum
from tamr_toolbox.data_io.df_connect import jdbc_info
from tamr_toolbox.models.data_type import JsonDict
from tamr_toolbox.data_io.file_system_type import FileSystemType
from tamr_unify_client.auth import UsernamePasswordAuth
from typing import Dict, List, Union, Optional

LOGGER = logging.getLogger(__name__)


@dataclass
class Client:
    """A data class for interacting with df_connect via jdbc.

    Args:
        host: the host where df_connect is running
        port: the port on which df_connect is listening
        protocol: http or https
        base_path: if using nginx-like proxy this is the redirect path
        tamr_username: the tamr account to use
        tamr_password: the password for the tamr account to use
        jbdc_info: configuration information for the jdbc connection
        cert: optional path to a certfile for authentication
    """

    host: str
    port: str
    protocol: str
    base_path: str
    tamr_username: str
    tamr_password: str
    jdbc_info: jdbc_info.JdbcInfo
    cert: Optional[str]


def from_config(
    config: JsonDict, config_key: str = "df_connect", jdbc_key: str = "ingest"
) -> Client:
    """Constructs a Client object from a json dictionary.

    Args:
        config: A json dictionary of configuration values
        config_key: block of the config to parse for values. Defaults to 'df_connect'
        jdbc_key: the key used to specify which block of df_connect-->jdbc in configuration to be
            used for picking up database connection information. Defaults to 'ingest'

    Returns:
        A Client object
    """

    # proxy and port redirect are optional
    base_path = config[config_key].get("base_path", "")
    port = config[config_key].get("port", "")

    # Optional cert may or may not be present in config file (back-compat from TBOX-295)
    if "cert" in config[config_key].keys():
        cert = config[config_key]["cert"]
    else:
        cert = None

    return Client(
        host=config[config_key]["host"],
        port=port,
        protocol=config[config_key]["protocol"],
        base_path=base_path,
        tamr_username=config[config_key]["tamr_username"],
        tamr_password=config[config_key]["tamr_password"],
        jdbc_info=jdbc_info.from_config(config, config_key=config_key, jdbc_key=jdbc_key),
        cert=cert,
    )


def create(
    *,
    host: str,
    port: str = "",
    protocol: str,
    base_path="",
    tamr_username: str,
    tamr_password: str,
    jdbc_dict: JsonDict,
    cert: Optional[str] = None,
) -> Client:
    """
    Simple wrapper for creating an instance of `Client` dataclass object.

    Args:
        host: the host where df_connect is running
        port: the port on which df_connect is listening
        protocol: http or https
        base_path: if using nginx-like proxy this is the redirect path
        tamr_username: the tamr account to use
        tamr_password: the password for the tamr account to use
        jdbc_dict: configuration information for the jdbc connection
        cert: optional path to a certfile for authentication

    Returns:
        An instance of `tamr_toolbox.data_io.df_connect.Client`
    """
    jdbc_information = jdbc_info.create(**jdbc_dict)
    return Client(
        host=host,
        port=port,
        protocol=protocol,
        base_path=base_path,
        tamr_username=tamr_username,
        tamr_password=tamr_password,
        jdbc_info=jdbc_information,
        cert=cert,
    )


def _get_url(connect_info: Client, api_path: str) -> str:
    """Constructs and returns url for request to df_connect. Valid for both http and https.

     Args:
         connect_info: An object of type JdbcInfo from which to pull host/port
         api_path: The api endpoint to be called

    Returns:
        url: A string of the request url formatted correctly for that instance of df_connect.
    """
    # handle port:
    if connect_info.port != "":
        port = ":" + connect_info.port
    else:
        port = connect_info.port
    # handle ssl_redirect
    api_path = "/".join([connect_info.base_path, api_path.lstrip("/")])
    api_path = api_path if api_path.startswith("/") else ("/" + api_path)
    if connect_info.protocol == "https":
        url = "https://" + connect_info.host + port + api_path
    else:
        url = "http://" + connect_info.host + port + api_path
    return url


def _get_query_config(jdbc_info: jdbc_info.JdbcInfo) -> Dict:
    """Packages configuration info into relevant query configuration json (dict) which is used
    for multiple df_connect API calls.

    Args:
        jdbc_info: JdbcInfo object from which to construct the query configuration.

    Returns:
        query_config: A dictionary suitable for usage in all df_connect API calls needing a
            queryConfig parameter
    """
    query_config = {
        "jdbcUrl": jdbc_info.jdbc_url,
        "dbUsername": jdbc_info.db_user,
        "dbPassword": jdbc_info.db_password,
        "fetchSize": jdbc_info.fetch_size,
    }
    return query_config


def _get_export_config(
    multi_value_delimiter: str = "|",
    limit_records: int = 0,
    columns_exclude_regex: str = "",
    rename_fields: Dict[str, str] = {},
) -> Dict:
    """Packages relevant pieces of JdbcExportInfo object into an exportDataConfig for jdbc export
     in form of json dictionary

    Args:
        multi_value_delimiter: value with which to delimit multivalues. default is |
        limit_records: number of records to stream. default is 0 (export all records)
        columns_exclude_regex: override config file for columnsExcludeRegex, default is empty
            string
        rename_fields: Dictionary in the format {“field_to_be_renamed”:“new_name”}

    Returns:
        A dictionary suitable for usage in all df_connect API calls around jdbc export
    """

    # build json object
    export_config = {
        "mergedArrayValuesDelimiter": multi_value_delimiter,
        "limitRecords": limit_records,
        "columnsExcludeRegex": columns_exclude_regex,
        "renameFields": rename_fields,
    }
    return export_config


def _get_avro_url_export_config(url: str, dataset_name: str) -> JsonDict:
    """Generates json needed for using avro(schema) export endpoints.
    Primary key is always set to [] since needing it is an artifact of the df-connect endpoint

    Args:
        url: the url to which the avro(schema) will be written
        dataset_name: the dataset for which the avro(schema) will be written
    """
    return {"url": url, "datasetName": dataset_name, "primaryKey": []}


def get_connect_session(connect_info: Client) -> requests.Session:
    """Returns an authenticated session using Tamr credentials from configuration.
    Raises an exception if df_connect is not installed or running correctly.

    Args:
        connect_info: An instance of a Client object

    Returns:
        An authenticated session

    Raises:
        RuntimeError: if a connection to df_connect cannot be established
    """
    auth = UsernamePasswordAuth(connect_info.tamr_username, connect_info.tamr_password)
    s = requests.Session()
    s.auth = auth
    s.headers.update({"Content-type": "application/json"})
    s.headers.update({"Accept": "application/json"})
    s.cert = connect_info.cert

    # test that df_connect is running properly
    url = _get_url(connect_info, "/api/service/health")
    try:
        r = s.get(url)
        r.raise_for_status()
    except (requests.exceptions.ConnectionError, requests.exceptions.HTTPError):
        raise RuntimeError(
            f"Tamr auxiliary service df-df_connect is either not running or not healthy at {url}!"
            f" Did you install it? Df-connect does not come with default Tamr installation."
            f" Check its status and your configuration."
        )
    return s


def ingest_dataset(
    connect_info: Client, *, dataset_name: str, query: str, primary_key=None
) -> JsonDict:
    """
    Ingest a dataset into Tamr via df-df_connect given dataset name, query string, and optional
    list of columns for primary key

    Args:
        dataset_name: Name of dataset
        query: jdbc query to execute in the database and results of which will be loaded into Tamr
        connect_info: A Client object for establishing session and loading jdbc parameters
        primary_key: list of columns to use as primary key. If None then df_connect will generate
            its own primary key

    Returns:
        JSON response from API call

    Raises:
        HTTPError: if the call to ingest the dataset was unsuccessful
    """
    # handle primary key
    if primary_key is None:
        primary_key = []
    else:
        primary_key = primary_key.split(",")

    # establish a df_connect session
    connect_session = get_connect_session(connect_info)

    # ingest data
    api_path = "/api/jdbcIngest/ingest"
    query_config = _get_query_config(connect_info.jdbc_info)
    ingest_data = {
        "query": query,
        "datasetName": dataset_name,
        "primaryKey": primary_key,
        "queryConfig": query_config,
    }
    ingest_url = _get_url(connect_info, api_path)
    LOGGER.info(
        f"Streaming data from {connect_info.jdbc_info.jdbc_url} to "
        f"Tamr with the following query: \n\t{query}"
    )
    r = connect_session.post(ingest_url, data=json.dumps(ingest_data))

    # check if successful and if so return True
    r.raise_for_status()
    return r.json()


def export_dataset(
    connect_info: Client,
    *,
    dataset_name: str,
    target_table_name: str,
    truncate_before_load: bool = False,
    **kwargs,
) -> JsonDict:
    """Export a dataset via jdbc to a target database.

    Args:
        dataset_name: the name of the dataset to export
        target_table_name: the table in the database to update
        truncate_before_load: whether or not to truncate the database table before load
        connect_info: A Client object for establishing session and loading jdbc parameters
        jdbc_key: the key for picking up relevant block for export from config file.
            See examples directory for usage

    Returns:
        JSON response from API call

    Raises:
        HTTPError: if the call to export the dataset was unsuccessful
    """

    # establish a df_connect session
    connect_session = get_connect_session(connect_info)

    # export data
    api_path = "/api/urlExport/jdbc"
    query_config = _get_query_config(connect_info.jdbc_info)
    export_data_config = _get_export_config(**kwargs)
    export_data = {
        "unifyDatasetName": dataset_name,
        "queryConfig": query_config,
        "exportDataConfig": export_data_config,
        "truncateBeforeLoad": truncate_before_load,
        "targetTableName": target_table_name,
    }

    export_url = _get_url(connect_info, api_path)
    LOGGER.info(
        f"Streaming data to {connect_info.jdbc_info.jdbc_url} from this "
        f"Tamr dataset: \n\t{dataset_name}"
    )
    r = connect_session.post(export_url, data=json.dumps(export_data))
    r.raise_for_status()
    return r.json()


def execute_statement(connect_info: Client, statement: str) -> JsonDict:
    """Calls the execute statement endpoint of df-df_connect.

    Args:
        statement: the SQL statement to be executed
        connect_info: A Client object for establishing session and loading jdbc parameters

    Returns:
        JSON response from API call

    Raises:
        HTTPError: if the call to df_connect was unsuccessful
    """

    # establish a df_connect session
    connect_session = get_connect_session(connect_info)

    # parse query config from main configuration
    query_config = _get_query_config(connect_info.jdbc_info)

    # export data
    api_path = "/api/jdbcIngest/execute"
    execute_data = {"queryConfig": query_config, "statement": statement}
    execute_url = _get_url(connect_info, api_path)
    LOGGER.info(
        f"Execute statement {statement} using the following jdbc url {query_config['jdbcUrl']}"
    )
    r = connect_session.post(execute_url, data=json.dumps(execute_data))
    r.raise_for_status()
    return r.json()


def profile_query_results(
    connect_info: Client, *, dataset_name: str, queries: List[str]
) -> JsonDict:
    """
    Profile the contents of JDBC queries via df_connect and write results to a Tamr dataset.
    For example the query "select * from table A" means that all rows from table A will be
    profiled, while "select * from table A where name=="my_name"" will only profile rows meeting
    the name=="my_name" condition.
    The same Tamr dataset can be used for profile results from multiple queries

    Args:
        dataset_name: Name of Tamr dataset for the profiling results
        queries: list of JDBC queries to execute in the database, the results of which will be
            profiled
        connect_info: A Client object for establishing session and loading jdbc parameters

    Returns:
        JSON response from API call

    Raises:
        HTTPError: if the call to profile the dataset was unsuccessful
    """
    # primary key must be passed but isn't used
    primary_key = []

    # establish a df_connect session
    connect_session = get_connect_session(connect_info)

    # run profiling
    api_path = "/api/jdbcIngest/profile"
    query_config = _get_query_config(connect_info.jdbc_info)
    profile_data = {
        "queryTargetList": [
            {"query": query, "datasetName": dataset_name, "primaryKey": primary_key}
            for query in queries
        ],
        "queryConfig": query_config,
    }
    profile_url = _get_url(connect_info, api_path)
    LOGGER.info(
        f"Profiling data from {connect_info.jdbc_info.jdbc_url} to Tamr with the "
        f"following queries: \n\t{queries}"
    )
    r = connect_session.post(profile_url, data=json.dumps(profile_data))

    # check if successful, and if so, return request JSON
    r.raise_for_status()
    return r.json()


def export_dataset_avro_schema(
    connect_info: Client, *, url: str, dataset_name: str, fs_type: Enum
) -> Union[JsonDict, bool]:
    """
    Takes a connect info object and writes the avro schema to specified url for specified
    dataset. By default assumes HDFS but if local_fs is set to true writes to server file system.

    Args:
        connect_info: The connect client to use
        url: the location in the relevant file system to which to write the avro schema
        dataset_name: the name of the dataset
        fs_type: the remote filesystem type. Currently supports 'HDFS' and 'LOCAL'

    Returns:
        json returned by df-connects /urlExport/<hdfs/serverfs>/avroSchema endpoint

    Raises:
        HTTPError: if the call to export the schema was unsuccessful
    """

    if fs_type == FileSystemType.LOCAL:
        api_path = f"/api/urlExport/{fs_type.value}/avroSchema"
    elif fs_type == FileSystemType.HDFS:
        api_path = f"/api/urlExport/{fs_type.value}/avroSchema"
    else:
        error = (
            f"trying to use unsupported type {fs_type}, supported are "
            f"'{FileSystemType.LOCAL.value}' and '{FileSystemType.HDFS.value}'"
        )
        LOGGER.error(error)
        raise ValueError(error)

    url_export_config = _get_avro_url_export_config(url, dataset_name)

    # establish a df_connect session and make API call
    connect_session = get_connect_session(connect_info)
    url = _get_url(connect_info, api_path)
    r = connect_session.post(url, data=json.dumps(url_export_config))
    r.raise_for_status()
    return r.json()


def export_dataset_as_avro(
    connect_info: Client, *, url: str, dataset_name: str, fs_type: Enum
) -> Union[JsonDict, bool]:
    """
    Takes a connect info object and writes the corresponding avro file to specified url for
    specified dataset. By default assumes HDFS but if local_fs is set to true writes to server
    file system.

    Args:
        connect_info: The connect client to use
        url: the location in the relevant file system to which to write the avro schema
        dataset_name: the name of the dataset
        fs_type: the remote filesystem type. Currently supports 'HDFS' and 'LOCAL'

    Returns:
        json returned by df-connects /urlExport/<hdfs/serverfs>/avroSchema endpoint

    Raises:
        ValueError: if using an unsupported type of file system
        HTTPError: if the call to export the dataset was unsuccessful
    """

    if fs_type == FileSystemType.LOCAL:
        api_path = f"/api/urlExport/{fs_type.value}/avro"
    elif fs_type == FileSystemType.HDFS:
        api_path = f"/api/urlExport/{fs_type.value}/avro"
    else:
        error = (
            f"trying to use unsupported type {fs_type}, supported are "
            f"'{FileSystemType.LOCAL.value}' and '{FileSystemType.HDFS.value}'"
        )
        LOGGER.error(error)
        raise ValueError(error)

    url_export_config = _get_avro_url_export_config(url, dataset_name)

    # establish a df_connect session and make API call
    connect_session = get_connect_session(connect_info)
    url = _get_url(connect_info, api_path)
    r = connect_session.post(url, data=json.dumps(url_export_config))
    r.raise_for_status()
    return r.json()



"""Tasks related to translating data with the google translation API"""
import html
import logging
import os
import time
from typing import Dict, List, Optional

from tamr_toolbox.enrichment.dictionary import TranslationDictionary

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke doc_src` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    from google.cloud.translate_v2 import Client as GoogleTranslateClient

LOGGER = logging.getLogger(__name__)


def _check_valid_translation_language(
    client: "GoogleTranslateClient", language: str, *, target_language: Optional[str] = None
) -> None:
    """
    Checks that the provided language is an accepted google translation api language with an
    option to specify the target language to check the source to target language combination
    is supported

    Args:
        client: a google translate api client
        language: the language to check
        target_language: the target language to translate to

    Returns:

    """
    languages = client.get_languages(target_language=target_language)
    valid_source_languages = [language["language"] for language in languages]
    if language not in valid_source_languages:
        if language != "auto":
            if target_language is None:
                error_message = (
                    f"Specified language {language} is not supported by the Google Translation API"
                    f"Valid languages are: {valid_source_languages}"
                )
            else:
                error_message = (
                    f"Translation from {language} to {target_language} is not supported by the "
                    "Google Translation API. "
                    f"Valid source languages for {target_language} are: {valid_source_languages}"
                )
            LOGGER.error(error_message)
            raise ValueError(error_message)


def _check_valid_translation_languages(
    client: "GoogleTranslateClient", source_language: str, target_language: str
) -> None:
    """
    Checks that the provided target and source language combination is an accepted translation
    with the google translation api

    Args:
        client: a google translate api client
        source_language: the target language to translate from
        target_language: the target language to translate to

    Returns:

    """
    if target_language == "auto":
        error_message = "'auto' is not a valid target language for translation"
        LOGGER.error(error_message)
        raise ValueError(error_message)
    else:
        _check_valid_translation_language(client, target_language)

    if source_language == "auto":
        LOGGER.info(
            "Source language is set to 'auto', "
            "the Google Translation API will automatically detect the source language"
        )
    else:
        _check_valid_translation_language(client, source_language, target_language=target_language)


def translation_client_from_json(json_credential_path: str) -> "GoogleTranslateClient":
    """
    Returns a Google translation client based on credentials stored in a Google credential
    json file

    Args:
        json_credential_path: path to the google credential json file

    Returns:
        A Google Translate Client

    """
    from google.cloud.translate_v2 import Client as GoogleTranslateClient

    LOGGER.info("Connecting to Google Translation Client")
    google_client = GoogleTranslateClient.from_service_account_json(json_credential_path)
    return google_client


def translate(
    phrases_to_translate: List[str],
    client: "GoogleTranslateClient",
    *,
    source_language: str = "auto",
    target_language: str = "en",
    translation_model: str = "nmt",
    num_of_tries: int = 4,
) -> Optional[Dict[str, TranslationDictionary]]:
    """
    Translate a list of text to a target language using google's translation api

    Args:
        phrases_to_translate: list of phrases to translate from the source language to the
            target language
        client: location of the credentials JSON read by the google_api client
        source_language: the language the text to translate is in, "auto" means the api_client
            google_api api_client will try to detect the source language automatically
        target_language: the language to translate into
        translation_model: google_api api_client api_client model to use, "nmt" or "pbmt".
            Choose "pbmt" if an "nmt" model doesn't exists for your source to target language pair
        num_of_tries: number of times to try to translate if the translation call fails

    Returns:
        A toolbox translation dictionary.
        None if the translation failed
    """
    _check_valid_translation_languages(
        client=client, target_language=target_language, source_language=source_language
    )

    if source_language == "auto":
        source_language = None

    num_attempts = 1
    last_attempt = False
    while num_attempts <= num_of_tries:
        if num_attempts == num_of_tries and num_of_tries > 1:
            LOGGER.warning(
                "WARNING: Failed to translate current chunk of phrases %s times. Final try.",
                num_of_tries - 1,
            )
            last_attempt = True
        try:
            response = client.translate(
                target_language=target_language,
                source_language=source_language,
                model=translation_model,
                values=phrases_to_translate,
            )
            if source_language is None:
                returned_translation = {
                    translation["input"]: TranslationDictionary(
                        standardized_phrase=translation["input"],
                        translated_phrase=html.unescape(translation["translatedText"]),
                        detected_language=translation["detectedSourceLanguage"],
                    )
                    for translation in response
                }
            else:
                returned_translation = {
                    translation["input"]: TranslationDictionary(
                        standardized_phrase=translation["input"],
                        translated_phrase=html.unescape(translation["translatedText"]),
                    )
                    for translation in response
                }
            LOGGER.debug(returned_translation)
            return returned_translation

        # TODO: check which exception returns the User Rate Limit error
        #  to better handle the exceptions
        except Exception as excp:
            if "User Rate Limit Exceeded" in str(excp) and not last_attempt:
                LOGGER.warning(
                    "Google api_client API user rate limit exceeded, "
                    "waiting 10 seconds and retrying."
                )
                time.sleep(10)
                num_attempts += 1
                continue
            else:
                error_message = f"Could not translate current chunk of phrases. Error: {excp}"
                LOGGER.error(error_message)
                num_attempts += 1
                continue

    LOGGER.error("Ran out of number of tries. Skipping.")
    return None



"""Tasks related to running jobs for Tamr Categorization projects"""
import logging
from typing import List

from tamr_unify_client.categorization.project import CategorizationProject
from tamr_unify_client.operation import Operation

from tamr_toolbox.models.project_type import ProjectType
from tamr_toolbox.utils import operation

LOGGER = logging.getLogger(__name__)


def _run_custom(
    project: CategorizationProject,
    *,
    run_update_unified_dataset: bool = False,
    run_apply_feedback: bool = False,
    run_update_results: bool = False,
    process_asynchronously: bool = False,
) -> List[Operation]:
    """Executes specified steps of a categorization project.

    Args:
        project: The target categorization project
        run_update_unified_dataset: Whether refresh should be called on the unified dataset
        run_apply_feedback: Whether train should be called on the pair matching model
        run_update_results: Whether predict should be called on the pair matching model
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        Operations that were run

    Raises:
        TypeError: if `project` is not a categorization project
    """
    if ProjectType[project.type] != ProjectType.CATEGORIZATION:
        error_msg = f"Cannot use as a categorization project. Project type: {project.type}"
        LOGGER.error(error_msg)
        raise TypeError(error_msg)
    else:
        project = project.as_categorization()

    completed_operations = []
    if run_update_unified_dataset:
        LOGGER.info(
            f"Updating the unified dataset for project {project.name} (id={project.resource_id})."
        )

        op = project.unified_dataset().refresh(asynchronous=process_asynchronously)

        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)

    if run_apply_feedback:
        LOGGER.info(
            f"Applying feedback to the categorization model for project {project.name} "
            f"(id={project.resource_id})."
        )
        op = project.model().train(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)

    if run_update_results:
        LOGGER.info(
            f"Updating categorization results for project {project.name} "
            f"(id={project.resource_id})."
        )
        op = project.model().predict(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)

    return completed_operations


def run(
    project: CategorizationProject,
    *,
    run_apply_feedback: bool = False,
    process_asynchronously: bool = False,
) -> List[Operation]:
    """Run the project

    Args:
        project: The target categorization project
        run_apply_feedback: Whether train should be called on the categorization model
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=True,
        run_apply_feedback=run_apply_feedback,
        run_update_results=True,
        process_asynchronously=process_asynchronously,
    )


def update_unified_dataset(
    project: CategorizationProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Updates the unified dataset for a categorization project

    Args:
        project: Target categorization project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=True,
        run_apply_feedback=False,
        run_update_results=False,
        process_asynchronously=process_asynchronously,
    )


def apply_feedback(
    project: CategorizationProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Trains the model only.

    Args:
        project: Target categorization project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_apply_feedback=True,
        run_update_results=False,
        process_asynchronously=process_asynchronously,
    )


def apply_feedback_and_update_results(
    project: CategorizationProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Trains the model and updates the categorization predictions of a categorization project

    Args:
        project: Target categorization project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_apply_feedback=True,
        run_update_results=True,
        process_asynchronously=process_asynchronously,
    )


def update_results_only(
    project: CategorizationProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Updates the categorization predictions based on the existing model of a
    categorization project

    Args:
        project: Target categorization project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow


    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_apply_feedback=False,
        run_update_results=True,
        process_asynchronously=process_asynchronously,
    )



"""Tasks related to managing unix-level operations like file and directory management"""

import logging
import os
import shutil
from datetime import datetime, timedelta

from typing import List, Union
from pathlib import Path

LOGGER = logging.getLogger(__name__)


def create_directories(
    directories: List[Union[Path, str]],
    *,
    require_absolute_path: bool = True,
    failure_if_exists: bool = False,
) -> None:
    """Creates directories recursively from a list of target directories if not already created

    Args:
        directories:  List of directories to create.
        require_absolute_path: Boolean flag indicating whether to require absolute paths.
        failure_if_exists: Boolean flag requiring failure if attempt to create a directory
            that already exists.
    Returns:
        None
    """
    for directory in directories:
        if not isinstance(directory, Path):
            directory = Path(directory)

        if require_absolute_path and not directory.is_absolute():
            raise ValueError(f"Expected absolute path for creation but received {directory}")

        if os.path.exists(directory):
            message = f"Tried creating {directory} but directory already exists"
            if failure_if_exists:
                raise OSError(message)
            LOGGER.warning(message)
        else:
            LOGGER.info(f"Creating directory {directory}")
            os.makedirs(directory)


def remove_directories(
    directories: List[Union[Path, str]],
    *,
    require_absolute_path: bool = True,
    failure_if_not_exists: bool = False,
    allow_recursive_deletes: bool = False,
) -> None:
    """Removes directories from a list of target directories if created.
    The directory must be empty

    Args:
        directories:  List of directories to remove
        require_absolute_path: Boolean flag indicating whether to require absolute paths.
        failure_if_not_exists: Boolean flag requiring failure if attempt to remove a directory
            that does not exist.
        allow_recursive_deletes: Boolean flag indicating whether to allow recursive deletion.
    Returns:
        None
    """
    LOGGER.warning(f"{directories}")
    for directory in directories:
        if not isinstance(directory, Path):
            directory = Path(directory)

        if require_absolute_path and not directory.is_absolute():
            raise ValueError(f"Expected absolute path for removal but received {directory}")

        if not os.path.exists(directory):
            message = f"Tried removing {directory} but directory does not exist"
            if failure_if_not_exists:
                raise OSError(message)
            LOGGER.warning(message)
        else:
            if allow_recursive_deletes:
                LOGGER.info(f"Removing directory {directory} recursively.")
                shutil.rmtree(directory)
            else:
                LOGGER.info(f"Removing directory {directory}")
                os.rmdir(directory)


def delete_old_files(
    data_path: Union[Path, str],
    *,
    num_days_to_keep: int,
    exclude_paths: List[Union[Path, str]] = None,
) -> List[str]:
    """
    Find all files in the directory data_path and its subdirectory which have not been modified for
    a specified number of days and delete them except if their subdirectory is specified in the
    folders to exclude

    Args:
        data_path: path to the root folder from where to search for files
        num_days_to_keep: number of days to keep files for
        exclude_paths: list of path to exclude from the search

    Returns:
        A list of deleted files

    Raises:
        ValueError: if num_days_to_keep is less than 0
        FileNotFoundError: if data_path doesn't exist
    """
    if num_days_to_keep < 0:
        value_error_message = (
            f"Argument for num_days_to_keep must be greater than 0. Found {num_days_to_keep}"
        )
        LOGGER.error(value_error_message)
        raise ValueError(value_error_message)
    # convert Path to str, to use with os.walk
    exclude_paths_strings = [str(p) for p in exclude_paths] if exclude_paths is not None else []
    if not os.path.exists(data_path):
        message = f"directory does not exist: {data_path}"
        LOGGER.error(message)
        raise FileNotFoundError(message)

    time_threshold = datetime.now() - timedelta(days=num_days_to_keep)
    files_to_delete = []
    for directory_path, _, files in os.walk(data_path):
        if directory_path not in exclude_paths_strings:
            for file in files:
                file_path = os.path.join(directory_path, file)
                if datetime.fromtimestamp(os.path.getmtime(file_path)) < time_threshold:
                    files_to_delete.append(file_path)

    LOGGER.info(
        f"Found {len(files_to_delete)} older than the retention policy of {num_days_to_keep} days"
    )
    LOGGER.info(f"Removing files older than the retention policy")
    for file in files_to_delete:
        LOGGER.debug(f"Deleting {file}")
        os.remove(file)

    return files_to_delete



"""Enum representing plan status"""
from enum import Enum
from tamr_toolbox.workflow.concurrent.PlanNodeStatus import PlanNodeStatus


class PlanStatus(Enum):
    """A class representing the status of a tamr_toolbox.workflow.concurrent.Planner plan"""

    PLANNED = "planned"
    RUNNING = "running"
    FAILED = "failed"
    SUCCEEDED = "succeeded"


def from_planner(planner) -> PlanStatus:
    """
    Create a PlanStatus object from a planner object
    Args:
        planner: the planner to pass in

    Returns:
        the plan's status  as a PlanStatus object
    """

    plan = planner.plan

    # check for failures first and make sure nothing can be run or is running
    if any(v.status == PlanNodeStatus.FAILED for k, v in plan.items()) and not any(
        (v.status == PlanNodeStatus.RUNNABLE or v.status == PlanNodeStatus.RUNNING)
        for k, v in plan.items()
    ):
        return PlanStatus.FAILED
    # then check to see if any job is running
    elif any(v.status == PlanNodeStatus.RUNNING for k, v in plan.items()):
        return PlanStatus.RUNNING
    # if not failed or running check if successful
    elif all(
        (v.status == PlanNodeStatus.SUCCEEDED) or (v.status == PlanNodeStatus.SKIPPABLE)
        for k, v in plan.items()
    ):
        return PlanStatus.SUCCEEDED
    # else assume still in planned state
    else:
        return PlanStatus.PLANNED



"""Tasks related to running jobs for Tamr Schema Mapping projects"""
from typing import List
import logging

from tamr_unify_client.operation import Operation
from tamr_unify_client.project.resource import Project

from tamr_toolbox.models.project_type import ProjectType
from tamr_toolbox.utils import operation

LOGGER = logging.getLogger(__name__)


def _run_custom(
    project: Project, *, run_update_unified_dataset=False, process_asynchronously: bool = False
) -> List[Operation]:
    """Executes specified steps of a schema mapping project.

    Args:
        project: Target schema mapping project
        run_update_unified_dataset: Whether refresh should be called on the unified dataset
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run

    Raises:
        TypeError: if the `project` is not a Schema Mapping projectgit
    """
    if ProjectType[project.type] != ProjectType.SCHEMA_MAPPING_RECOMMENDATIONS:
        error_msg = f"Cannot use as a schema mapping project. Project type: {project.type}"
        LOGGER.error(error_msg)
        raise TypeError(error_msg)

    completed_operations = []
    if run_update_unified_dataset:
        LOGGER.info(
            f"Updating the unified dataset for project {project.name} (id={project.resource_id})."
        )
        op = project.unified_dataset().refresh(asynchronous=process_asynchronously)

        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)

    return completed_operations


def run(project: Project, *, process_asynchronously: bool = False) -> List[Operation]:
    """Run the project

    Args:
        project: Target schema mapping project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project, run_update_unified_dataset=True, process_asynchronously=process_asynchronously
    )


def update_unified_dataset(
    project: Project, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Updates the unified dataset for a schema mapping project

    Args:
        project: Target schema mapping project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project, run_update_unified_dataset=True, process_asynchronously=process_asynchronously
    )



"""Tasks related to logging within scripts"""
import os
import sys
import logging
import datetime
from typing import Optional


def _get_log_filename(log_prefix: str = "", date_format: str = "%Y-%m-%d") -> str:
    """Generate standard format log-file names.

    Args:
        log_prefix:  prefix for log filename
        date_format: the date format to be used in the file name

    Returns:
        filename in format {log_prefix}_{current_date}
    """

    # When log_prefix is populated, append an underscore and
    # ignore any trailing underscore provided by the user
    if log_prefix != "":
        log_prefix = f"{log_prefix.rstrip('_')}_"

    date = datetime.datetime.now().strftime(date_format)
    log_filename = f"{log_prefix}{date}.log"
    return log_filename


def _add_handler(logger: logging.Logger, log_directory: Optional[str] = None, **kwargs) -> None:
    """Adds a handler to a logger, either a logging.StreamHandler if log_directory is None
    otherwise a logging.FileHandler piped to the directory specified.

    Args:
        logger: the logging.Logger class to which you would like to add a handler
        log_directory: Optional log directory to pass. If not None a FileHandler is added,
            otherwise a StreamHandler
        **kwargs: Keyword arguments for the _get_log_filename
    """
    if log_directory is None:
        handler = logging.StreamHandler()
    else:
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)

        handler = logging.FileHandler(os.path.join(log_directory, _get_log_filename(**kwargs)))

    # for some reason you need to set both of these - setting to same value to avoid confusion
    logger.setLevel(logging.INFO)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(levelname)s <%(thread)d> [%(asctime)s] %(name)s <%(filename)s:%(lineno)d> %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def create(
    name: str,
    *,
    log_to_terminal: bool = True,
    log_directory: Optional[str] = None,
    log_prefix: str = "",
    date_format: str = "%Y-%m-%d",
) -> logging.Logger:
    """Return logger object with pre-defined format.
    Log file will be located under log_directory with file name
    <log_prefix>_<date>.log, quashing extra separating underscores. Defaults to <date>.log.

    For use in scripts only. To log in module files, use the standard library `logging` module with
    a module-level logger and enable package logging.  See
    https://docs.python.org/3/howto/logging.html#advanced-logging-tutorial

    >>> log = logging.getLogger(__name__)

    Args:
        name: This sets the name of your logger instance. It does not affect the file name.
            To change the filename use log_prefix
        log_to_terminal: Boolean indicating whether or not to log messages to the terminal.
        log_directory: The directory to place log files inside
        log_prefix: The string to prepend to the date in the log file name.
        date_format: format string for date suffix on log file name
    Returns:
        Logger object
    """
    if type(name) is not str:
        raise TypeError
    if name == "":
        raise ValueError

    logger = logging.getLogger(name)
    if log_to_terminal:
        _add_handler(logger, log_prefix=log_prefix, date_format=date_format)
    if log_directory is not None:
        _add_handler(
            logger, log_directory=log_directory, log_prefix=log_prefix, date_format=date_format
        )

    # enable logging of uncaught exceptions
    def log_exception(exc_type, exc_value, exc_traceback) -> None:
        logger.error("Uncaught exception:", exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = log_exception

    return logger


def set_logging_level(logger_name: str, level: str) -> None:
    """A useful method for setting logging level for all a given logger and its handlers.

    Args:
        logger_name: the name of the logger for which to set the level
        level: log level to use. The set available from core logging package is 'debug', 'info',
            'warning', 'error'
    """
    log = logging.getLogger(logger_name)
    logging_level = logging.getLevelName(level.upper())
    log.setLevel(logging_level)
    for x in log.handlers:
        x.setLevel(logging_level)


def enable_package_logging(
    package_name: str,
    *,
    log_to_terminal: bool = True,
    log_directory: Optional[str] = None,
    level: Optional[str] = None,
    log_prefix: str = "",
    date_format: str = "%Y-%m-%d",
) -> None:
    """A helper function to enable package logging for any package following
    python best practices for logging names (i.e. logger name == package.module.submodule).

    Args:
        package_name: the name of the package for which to enable logging
        log_to_terminal: Boolean indicating whether or not to log messages to the terminal
        log_directory: optional log directory which the package will write logs
        level: optional level to specify, default is WARNING (inherited from base logging package)
        log_prefix: Optional prefix for log files, if None will be blank string
        date_format: Optional date format for log file
    """

    package_logger = logging.getLogger(package_name)
    _add_handler(package_logger, log_directory, log_prefix=log_prefix, date_format=date_format)
    if log_to_terminal:
        _add_handler(package_logger, log_prefix=log_prefix, date_format=date_format)

    if level is not None:
        set_logging_level(package_name, level)


def enable_toolbox_logging(
    *,
    log_to_terminal: bool = True,
    log_directory: Optional[str] = None,
    level: Optional[str] = None,
    log_prefix: str = "",
    date_format: str = "%Y-%m-%d",
) -> None:
    """A simple wrapper to enable_package_logging to give friendly call for users.

    Args:
        log_to_terminal: Boolean indicating whether or not to log messages to the terminal
        log_directory: optional directory to which to write tamr_toolbox logs
        level: Optional logging level to specify, default is WARNING
            (inherited from base logging package)
        log_prefix: Optional prefix for log files, if None will be blank string
        date_format: Optional date format for log file
    """

    enable_package_logging(
        "tamr_toolbox",
        log_to_terminal=log_to_terminal,
        log_directory=log_directory,
        level=level,
        log_prefix=log_prefix,
        date_format=date_format,
    )



"""Tasks related to attribute configurations as part of Tamr projects"""
from tamr_unify_client.project.attribute_configuration.resource import AttributeConfigurationSpec

from tamr_unify_client.project.resource import Project

import logging

LOGGER = logging.getLogger(__name__)


def _check_tokenizer(value):
    """Check tokenizer value is valid

    Args:
        value: Value of variable to check

    Raises:
        ValueError: Raised if value if not in allowed list
    """
    # List of allowed values
    allowed_values = ["DEFAULT", "STEMMING_EN", "BIGRAM", "TRIGRAM", "BI-WORD", "REGEX"]

    # Raise error if not in allowed list
    if value not in allowed_values:
        raise ValueError(
            f"Value '{value}' not valid for attribute configuration variable "
            + f"tokenizer. Allowed values: {allowed_values}"
        )


def _check_similarity_function(value):
    """Check similarityFunction value is valid

    Args:
        value: Value of variable to check

    Raises:
        ValueError: Raised if value if not in allowed list
    """
    # List of allowed values
    allowed_values = ["COSINE", "JACCARD", "ABSOLUTE_DIFF", "RELATIVE_DIFF"]

    # Raise error if not in allowed list
    if value not in allowed_values:
        raise ValueError(
            f"Value '{value}' not valid for attribute configuration variable "
            + f"similarityFunction. Allowed values: {allowed_values}"
        )


def _check_attribute_role(value):
    """Check attributeRole value is valid

    Args:
        value: Value of variable to check

    Raises:
        ValueError: Raised if value if not in allowed list
    """
    # List of allowed values
    allowed_values = ["CLUSTER_NAME_ATTRIBUTE", "SUM_ATTRIBUTE", ""]

    # Raise error if not in allowed list
    if value not in allowed_values:
        raise ValueError(
            f"Value '{value}' not valid for attribute configuration variable "
            + f"attributeRole. Allowed values: {allowed_values}"
        )


def _check_enabled_for_ml(value):
    """Check enableForMl value is valid

    Args:
        value: Value of variable to check

    Raises:
        ValueError: Raised if value if not in allowed list
    """
    # List of allowed values
    allowed_values = [True, False]

    # Raise error if not in allowed list
    if value not in allowed_values:
        raise ValueError(
            f"Value '{value}' not valid for attribute configuration variable "
            + f"enabledForMl. Allowed values: {allowed_values}"
        )


def get_attribute_configurations(project: Project) -> AttributeConfigurationSpec:
    """Store the attribute configurations in a list for a given project

    Args:
        project: Project containing attributes with configurations defined

    Returns:
        List of AttributeConfigurationSpec
    """
    # Get all the attribute configurations from a project
    attribute_configuration_all = project.attribute_configurations()

    # Itterate through all attributes and store in a list
    attribute_configuration_spec = [
        attribute_configuration.spec() for attribute_configuration in attribute_configuration_all
    ]

    return attribute_configuration_spec


def update_attribute_configuration(
    project: Project,
    attribute_name: str,
    attribute_role: str = None,
    similarity_function: str = None,
    enabled_for_ml: bool = None,
    tokenizer: str = None,
    numeric_field_resolution: float = None,
):
    """Update the attribute configuration variables of a given
    attribute in a project

    Args:
        project (Project): Project in which attribute is present
        attribute_name (str): Name of attribute to update configuration
        attribute_role (str, optional): The specific role, if any, of the attribute in the project.
            Defaults to None.
        similarity_function (str, optional): The similarity function to use for the unified dataset
            attribute. Defaults to None.
        enabled_for_ml (bool, optional): Whether the unified dataset attribute is being included in
            machine learning operations. Defaults to None.
        tokenizer (str, optional): The tokenizer used for tokenizing text values. Defaults to
            None.
        numeric_field_resolution (float, optional): Indicates how to process numeric values.
            Defaults to None.

    Raises:
        RuntimeError: If the attribute is not found in the project
        ValueError: If any of the optional attribute configuration values are not valid

    Returns:
        Response of API request to update attribute configuration
    """
    LOGGER = logging.getLogger(__name__)

    # Define variable for use later
    attribute_found = False
    attribute_variable_dict = {}

    # Check project has attribute present
    project_attribute_configuration_all = get_attribute_configurations(project=project)
    # Get the values from the configurations

    # Loop through all the attributes
    for attribute_config_single in project_attribute_configuration_all:
        # Get values from configuration spec
        attribute_config_single_dict = attribute_config_single.to_dict()
        # Check if name matches
        if attribute_config_single_dict["attributeName"] == attribute_name:
            # Set to true if match
            attribute_found = True
            # Save attribute info
            attribute_config_spec = attribute_config_single
            attribute_config = attribute_config_single_dict
    if attribute_found:
        LOGGER.info(f"Attribute {attribute_name} sucessfully found in project {project.name}")
        pass
    else:
        raise RuntimeError(f"Attribute {attribute_name} not in project {project.name}!")

    # Check there are fields to be updated
    variable_list = [
        attribute_role,
        similarity_function,
        enabled_for_ml,
        tokenizer,
        numeric_field_resolution,
    ]
    empty_variables = all(variable is None for variable in variable_list)
    if empty_variables:
        raise ValueError(
            "No configuration variables are specified. Enter a value for at least one"
            + "of the following variables: <attributeRole>, <similarityFunction>,"
            + "<enabledForMl>, <tokenizer>, <numericFieldResolution>!"
        )

    # Collect configuration variables to update
    if attribute_role is not None:
        _check_attribute_role(attribute_role)
        attribute_variable_dict["attributeRole"] = attribute_role
    if similarity_function is not None:
        _check_similarity_function(similarity_function)
        attribute_variable_dict["similarityFunction"] = similarity_function
    if enabled_for_ml is not None:
        _check_enabled_for_ml(enabled_for_ml)
        attribute_variable_dict["enabledForMl"] = enabled_for_ml
    if tokenizer is not None:
        _check_tokenizer(tokenizer)
        attribute_variable_dict["tokenizer"] = tokenizer
    if numeric_field_resolution is not None:
        attribute_variable_dict["numericFieldResolution"] = numeric_field_resolution

    # Loop through new config values and check for changes
    for key, value in attribute_variable_dict.items():
        # Check if an update
        if value == attribute_config[key]:
            LOGGER.warn(f"Attribute configuration variable {key} will not change!")
            pass
        else:
            LOGGER.info(
                f"Attribute configuration variable {key} will be updated "
                + f"from {attribute_config[key]} to {value}!"
            )
            # Update attribute configuration value which has changed
            attribute_config[key] = value

    # Send updates to Tamr
    new_attribute_configuration = attribute_config_spec.from_data(attribute_config)
    response = new_attribute_configuration.put()

    return response



"""Defines the package tamr_toolbox for user installations"""
from setuptools import setup, find_packages

with open("requirements.txt") as f:
    required = f.read().splitlines()

with open("version.txt") as f:
    full_version = f.read().splitlines()[0]
    version_number = full_version.split("-")[0]

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="tamr_toolbox",
    version=version_number,
    author="Tamr Inc.",
    author_email="",
    description="Tools for Tamr",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/Datatamer/tamr-toolbox",
    packages=find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.6",
    install_requires=required,
    extras_require={
        # Super set including all optional dependencies
        "all": [
            "pandas>=0.21.0",
            "aiohttp==3.7.4.post0",
            "slackclient==2.7.2",
            "responses==0.10.14",
            "paramiko>=2.8.0",
            "google-cloud-translate==3.7.4",
            "google-cloud-storage>=2.0.0",
            "boto3>=1.21.21",
            "boto3-stubs-lite[essential]>=1.21.21",
        ],
        # Individual sets of dependencies
        "address-validation": ["googlemaps==4.10.0"],
        "pandas": ["pandas>=0.21.0"],
        "slack": ["aiohttp==3.7.4.post0", "slackclient==2.7.2"],
        "testing": ["responses==0.10.14"],
        "translation": ["google-cloud-translate==3.7.4"],
        "ssh": ["paramiko>=2.8.0"],
        "gcs": ["google-cloud-storage>=2.0.0"],
        "s3": ["boto3>=1.21.21", "boto3-stubs-lite[essential]>=1.21.21"],
    },
)



"""Enum representing operation states in Tamr"""
from enum import Enum


class OperationState(Enum):
    """A dataclass representing job state codes in Tamr

    Args:
        SUCCEEDED: Status when a job has completed successfully
        FAILED: Status when a job has been terminated due to an error
        CANCELED: Status when a job been terminated due to cancelation by a user or the job service
        PENDING: Status when a job has been submitted and is waiting to begin running
        RUNNING: Status when a job is active
    """

    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    CANCELED = "CANCELED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"



"""Tasks related to creation of Email notifications"""
from typing import Union, List, Optional, Callable
from tamr_toolbox.models.data_type import JsonDict
from tamr_unify_client import Client
from tamr_unify_client.operation import Operation
from tamr_toolbox.utils.operation import monitor

from tamr_toolbox.models.operation_state import OperationState
from tamr_toolbox.utils.operation import from_resource_id


def _monitor_job(
    tamr: Client,
    *,
    send_message: Callable[..., JsonDict],
    send_status_function: Callable[..., JsonDict],
    operation: Union[int, str, Operation],
    poll_interval_seconds: float = 1,
    timeout_seconds: Optional[float] = None,
    notify_states: Optional[List[OperationState]] = None,
    **kwargs,
) -> List[JsonDict]:
    """Monitors a Tamr Operation and sends a message when the job status is updated

    Args:
        tamr: A Tamr client
        send_message: either send_email from emails or send_message function from slack
        send_status_function: _send_job_status_message function from
        notifications.slack or notifications.email
        operation: A job ID or a Tamr operation
        poll_interval_seconds: Time interval (in seconds) between subsequent polls
        timeout_seconds: Time (in seconds) to wait
        notify_states : States for which notifications should be sent, use None for all states

    Returns:
        A list of messages with their response codes
    """
    if notify_states is None:
        notify_states = [
            OperationState.SUCCEEDED,
            OperationState.FAILED,
            OperationState.CANCELED,
            OperationState.PENDING,
            OperationState.RUNNING,
        ]
    list_responses = []

    if isinstance(operation, Operation):
        op = operation
    else:
        op = from_resource_id(tamr=tamr, job_id=operation)

    # send message for initial state
    status = OperationState[op.state]
    if status in notify_states:
        response = send_status_function(operation=op, notify_states=notify_states, **kwargs)
        list_responses.append(response)

    while status not in [OperationState.SUCCEEDED, OperationState.FAILED, OperationState.CANCELED]:
        try:
            op = monitor(
                operation=op,
                poll_interval_seconds=poll_interval_seconds,
                timeout_seconds=timeout_seconds,
            )
            status = OperationState[op.state]
            # send message on state change
            response = send_status_function(operation=op, notify_states=notify_states, **kwargs)
            list_responses.append(response)
        except TimeoutError:
            timeout_message = (
                f"The job {op.resource_id}: {op.description} took longer "
                f"than {timeout_seconds} seconds to resolve."
            )
            response = send_message(
                message=timeout_message, subject_line=f"Job {op.resource_id}: Timeout", **kwargs
            )
            list_responses.append(response)
            return list_responses

    return list_responses



"""Tasks related to validation and refresh of address data using Google Maps API"""
import logging
import math
import os
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Tuple

from tamr_toolbox.enrichment.address_mapping import AddressValidationMapping, save, update
from tamr_toolbox.enrichment.api_client.google_address_validate import validate
from tamr_toolbox.enrichment.enrichment_utils import join_clean_tuple

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    from googlemaps import Client as GoogleMapsClient

LOGGER = logging.getLogger(__name__)


def get_addr_to_validate(
    input_addresses: List[Tuple[Optional[str], ...]],
    addr_mapping: Dict[str, AddressValidationMapping],
    expiration_date_buffer: timedelta = timedelta(days=1),
) -> List[str]:
    """Find addresses not previously validated or validated too long ago.

    Args:
        input_addresses: list of addresses to validate
        addr_mapping: a dict of address validation mapping data
        expiration_date_buffer: re-validate addresses if they are within this period of expiring

    Returns:
        List of standardized addresses not present as keys of the mapping dictionary

    Raises:
        ValueError: if negative `expiration_date_buffer` is supplied
    """

    if str(expiration_date_buffer)[0] == "-":
        raise ValueError("Buffer time for expiration date cannot be negative.")

    count_new_addr = 0
    count_stale_addr = 0
    addr_to_validate = []

    for addr in input_addresses:
        joined_addr = join_clean_tuple(addr)
        if joined_addr not in addr_mapping.keys():
            addr_to_validate.append(joined_addr)
            count_new_addr += 1
        elif addr_mapping[joined_addr].expiration < str(datetime.now() + expiration_date_buffer):
            addr_to_validate.append(joined_addr)
            count_stale_addr += 1

    LOGGER.info(
        "From %s sent for validation, %s have been not been validated before; %s are stale.",
        len(input_addresses),
        count_new_addr,
        count_stale_addr,
    )
    LOGGER.debug("Items to validate: %s", addr_to_validate)
    return addr_to_validate


def from_list(
    all_addresses: List[Tuple[Optional[str], ...]],
    client: "GoogleMapsClient",
    dictionary: Dict[str, AddressValidationMapping],
    *,
    region_code: Optional[str],
    enable_usps_cass: bool = False,
    intermediate_save_every_n: Optional[int] = None,
    intermediate_save_to_disk: bool = False,
    intermediate_folder: str = "/tmp",
    expiration_date_buffer: timedelta = timedelta(days=1)
) -> Dict[str, AddressValidationMapping]:
    """Validate a list of addresses.

    The validation is saved in a dictionary on your local file system before updating the
    main dictionary.

    Args:
        all_addresses: List of addresses to validate
        client: a googlemaps api client
        dictionary: a toolbox validation dictionary
        region_code: optional region code, e.g. 'US' or 'FR', to pass to the maps API
        enable_usps_cass: bool: whether to use USPS validation; only for 'US'/'PR' regions
        intermediate_save_every_n: save periodically api_client dictionary every n
            addresses validated; if not set, will save only at end of processing
        intermediate_save_to_disk: decide whether to save periodically the dictionary to disk to
            avoid loss of validation data if code breaks
        intermediate_folder: path to folder where dictionary will be save periodically to avoid
            loss of validation data
        expiration_date_buffer: re-validate addresses if they are within this period of expiring

    Returns:
        The updated validation dictionary
    """
    if intermediate_save_every_n == 0 or intermediate_save_every_n is None:
        intermediate_save_every_n = math.inf

    unique_all_addresses = list(set(all_addresses))
    nbr_of_unique_addresses = len(unique_all_addresses)

    addresses_to_validate = get_addr_to_validate(
        unique_all_addresses, dictionary, expiration_date_buffer=expiration_date_buffer
    )
    nbr_addresses_to_validate = len(addresses_to_validate)

    if nbr_addresses_to_validate == 0:
        LOGGER.info("All addresses to validate are found in the local dictionary.")

    else:
        LOGGER.info(
            "Of %s addresses to validate, %s were not found in the dictionary or were expired.",
            nbr_of_unique_addresses,
            nbr_addresses_to_validate,
        )

        tmp_dictionary = {}
        for idx, address in enumerate(addresses_to_validate):
            validated_address = validate(
                address_to_validate=address,
                client=client,
                locality=None,
                region_code=region_code,
                enable_usps_cass=enable_usps_cass,
            )
            if validated_address is not None:
                tmp_dictionary.update({address: validated_address})

            if ((idx + 1) % intermediate_save_every_n) == 0:
                LOGGER.info("Saving intermediate outputs")
                update(main_dictionary=dictionary, tmp_dictionary=tmp_dictionary)
                if intermediate_save_to_disk:
                    save(addr_mapping=dictionary, addr_folder=intermediate_folder)
                # Reset temporary results after saving
                tmp_dictionary = {}

        # update dictionary
        update(main_dictionary=dictionary, tmp_dictionary=tmp_dictionary)
        if intermediate_save_to_disk:
            save(addr_mapping=dictionary, addr_folder=intermediate_folder)

    return dictionary



import os

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    from google.cloud import storage
    from mypy_boto3_s3.client import S3Client


def gcs_upload(
    cloud_client: "storage.Client",
    *,
    source_filepath: str,
    destination_filepath: str,
    bucket_name: str,
):
    """Upload data to a Google storage bucket
    Args:
        source_filepath: file path of source file being uploaded
        destination_filepath: path to google bucket
        cloud_client: google storage client with user credentials
        bucket_name: name of google bucket
    """
    bucket = cloud_client.get_bucket(bucket_name)
    blob = bucket.blob(destination_filepath)
    blob.upload_from_filename(source_filepath)


def gcs_download(
    cloud_client: "storage.Client",
    *,
    source_filepath: str,
    destination_filepath: str,
    bucket_name: str,
):
    """Download data to a Google storage bucket
    Args:
        source_filepath: Path to file being downloaded on Google
        destination_filepath: Destination filepath for file being downloaded
        cloud_client: google storage client with user credentials
        bucket_name: name of google bucket being downloaded from
    """

    bucket = cloud_client.get_bucket(bucket_name)
    this_blob = bucket.blob(source_filepath)
    this_blob.download_to_filename(destination_filepath)


def s3_upload(
    cloud_client: "S3Client", *, source_filepath: str, destination_filepath: str, bucket_name: str
):
    """Upload data to Amazon AWS bucket
    Args:
        source_filepath: file path of source file being uploaded
        destination_filepath: path to AWS bucket
        cloud_client: AWS client with user credentials
        bucket_name: name of AWS bucket
        downloads copy of file being uploaded to filepath
    """
    cloud_client.upload_file(
        Filename=source_filepath, Bucket=bucket_name, Key=destination_filepath
    )


def s3_download(
    cloud_client: "S3Client", *, source_filepath: str, destination_filepath: str, bucket_name: str
):
    """Download data from an Amazon AWS bucket
    Args:
        source_filepath: Path to file being downloaded on AWS
        destination_filepath: Destination filepath for file being downloaded
        cloud_client: AWS client with user credentials
        bucket_name: name of AWS bucket being downloaded from
    """
    cloud_client.download_file(
        Bucket=bucket_name, Key=source_filepath, Filename=destination_filepath
    )



"""Tasks related to creation of Slack notifications"""
import logging
import os
from typing import Union, List, Optional
from tamr_toolbox.models.data_type import JsonDict

from tamr_unify_client import Client
from tamr_unify_client.operation import Operation
from tamr_toolbox.notifications.common import _monitor_job as monitor_job_common

from tamr_toolbox.models.operation_state import OperationState
from tamr_toolbox.utils.operation import get_details

LOGGER = logging.getLogger(__name__)

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    import slack


def send_message(
    *,
    slack_client: "slack.WebClient",
    channel: str,
    message: str,
    raise_error: bool = True,
    **kwargs,
) -> JsonDict:
    """Sends a message to a pre-defined Slack channel

    Args:
        slack_client: A Slack WebClient
        channel: A (public) Slack channel to receive messages
        message: A message to be sent to a Slack channel
        raise_error: A boolean value to opt out raising Slack API errors

    Returns:
        The data property of the Slack WebClient object

    Raises:
        SlackApiError: if the SlackClient does not post the message correctly
            and `raise_error` is set to True
    """

    # This function requires slack, an optional dependency
    from slack.errors import SlackApiError

    try:
        slack_response = slack_client.chat_postMessage(channel=channel, text=message)
    except SlackApiError as e:
        LOGGER.error(f"Error: {e}")

        if not raise_error:
            return {
                "type": "SlackError",
                "text": f'The message: "{message}" failed to send to channel: {channel}',
                "details": e.response["error"],
            }
        else:
            raise e

    else:
        return slack_response.data


def _send_job_status_message(
    *,
    slack_client: "slack.WebClient",
    channel: str,
    operation: Operation,
    notify_states: List[OperationState],
) -> JsonDict:
    """Checks operation state and if in `notify_states` sends the message.

    Args:
        slack_client: A Slack WebClient
        channel: A (public) Slack channel to receive messages
        operation: A Tamr Operation
        notify_states: States for which notifications should be sent

    Returns:
        The data property of the Slack WebClient object
    """
    state = OperationState[operation.state]
    if state in notify_states:
        message = get_details(operation=operation)
        return send_message(slack_client=slack_client, channel=channel, message=message)


def monitor_job(
    tamr: Client,
    *,
    slack_client: "slack.WebClient",
    channel: str,
    operation: Union[int, str, Operation],
    poll_interval_seconds: float = 1,
    timeout_seconds: Optional[float] = None,
    notify_states: Optional[List[OperationState]] = None,
) -> List[JsonDict]:
    """Monitors a Tamr Operation and sends a Slack message to a channel when the job status is updated

    Args:
        tamr: A Tamr client
        slack_client: A Slack WebClient
        channel: A (public) Slack channel to receive messages
        operation: A job ID or a Tamr operation
        poll_interval_seconds: Time interval (in seconds) between subsequent polls
        timeout_seconds: Time (in seconds) to wait
        notify_states : States for which notifications should be sent, use None for all states

    Returns:
        A list of messages with their config sent via the Slack WebClient object
    """
    list_responses = monitor_job_common(
        tamr=tamr,
        send_message=send_message,
        send_status_function=_send_job_status_message,
        slack_client=slack_client,
        channel=channel,
        operation=operation,
        poll_interval_seconds=poll_interval_seconds,
        timeout_seconds=timeout_seconds,
        notify_states=notify_states,
    )

    return list_responses



"""Tasks related to metrics for Tamr Categorization projects"""
import logging

from tamr_unify_client.categorization.taxonomy import Taxonomy
from tamr_unify_client.dataset.resource import Dataset
from tamr_unify_client.project.resource import Project

from tamr_toolbox.models import data_type
from tamr_toolbox.utils import operation

LOGGER = logging.getLogger(__name__)
CONFIDENCE_DATASET_SUFFIX = "classifications_average_confidences"


def _check_dataset_with_confidence(dataset: Dataset) -> None:
    """
    Checks if the dataset contains necessary attributes for obtaining confidence information

    Args:
        dataset: Tamr dataset object

    Returns:
        True if the dataset contains required attributes to compute the average confidence;
        False otherwise

    Raises:
        RuntimeError: if the dataset does not contain 'classificationPath' or 'averageConfidence'
        attributes
    """
    # check required attributes are present
    attribute_names = [attribute.name for attribute in dataset.attributes.stream()]
    if "classificationPath" not in attribute_names or "averageConfidence" not in attribute_names:
        wrong_attribute_error = (
            f"You might be using a version of Tamr unsupported for this functionality because "
            f" 'classificationPath' and 'averageConfidence' attributes are missing are missing"
            f" from dataset {dataset.name}."
        )
        LOGGER.error(wrong_attribute_error)
        raise RuntimeError(wrong_attribute_error)


def _check_taxonomy_depth(project: Project, *, tier: int) -> None:
    """
    Checks the maximum depth of the taxonomy associated wit

    Args:
        project: Tamr project object
        tier: integer specifying the tier from which to extract categories

    Returns:
        whether tier exceed the maximum taxonomy depth or not

    Raises:
        ValueError: if tier is greater than maximum taxonomy depth
    """

    # depth check is not required for leaf nodes
    if tier == -1:
        return

    max_depth = 0
    classification_project = project.as_categorization()
    taxonomy = classification_project.taxonomy()
    categories = taxonomy.categories()

    for category in categories:
        if len(category.path) > max_depth:
            max_depth = len(category.path)

    if max_depth < tier:
        invalid_tier_value_error = (
            f"Invalid value for tier {tier}. Maximum depth detected is {max_depth}."
        )
        LOGGER.error(invalid_tier_value_error)
        raise ValueError(invalid_tier_value_error)


def _create_leaf_node_set(taxonomy: Taxonomy) -> set:
    """
    Extracts leaf nodes from taxonomy and returns them in a set.
    For a taxonomy with paths [[cat1], [cat1, cat11], [cat2], [cat1, cat12], [cat1, cat11, cat111]]
     the function returns set('cat2', 'cat1|cat11|cat111', 'cat1|cat12')

    Args:
        taxonomy: Tamr Taxonomy object for a categorization project

    Returns:
        Set of all leaf nodes in taxonomy, where node paths are joined by '|' if taxonomy is
        multi-level
    """
    all_set = set()
    parent_set = set()
    for category in taxonomy.categories().stream():
        all_set.add("|".join(category.path))
        if category.parent():
            parent_set.add("|".join(category.parent().path))
    leaf_set = all_set - parent_set
    return leaf_set


def _extract_confidence(*, dataset: Dataset, category_set: set) -> data_type.JsonDict:
    """
    Extracts tier-specific average confidence from a Tamr internal dataset
    `<unified dataset name>_classifications_average_confidences` to a dictionary

    Args:
        dataset: Tamr internal Dataset with a name ending in
        `_unified_dataset_classifications_average_confidences`
        category_set: set of category paths at the desired tier

    Returns:
        dictionary - keys are category paths, joined by '|' if multi-level taxonomy. Values are
        average confidence of the corresponding keys, where it is None if no confidence exists for
        the category.
    """
    confidence_dict = {}
    for record in dataset.records():
        path = "|".join(record["classificationPath"])
        if path in category_set:
            confidence_dict[path] = record["averageConfidence"]

    empty_confidence_categories = category_set - set(confidence_dict.keys())
    for category in empty_confidence_categories:
        confidence_dict[category] = None

    return confidence_dict


def _get_categories_at_tier(project: Project, *, tier: int) -> set:
    """
    Extracts categories at tier from a taxonomy associated with Project

    Args:
        project: Tamr project object
        tier: integer specifying the tier to extract the categories;
              -1 will return all leaf categories

    Returns:
        set of category paths at tier, joined by '|' if multi-level taxonomy
    """
    classification_project = project.as_categorization()
    taxonomy = classification_project.taxonomy()
    categories = taxonomy.categories()

    category_set = set()
    if tier > 0:
        for category in categories:
            if len(category.path) == tier:
                category_set.add("|".join(category.path))
    else:
        # leaf nodes
        category_set = _create_leaf_node_set(taxonomy)
    return category_set


def _get_dataset_with_confidence(project: Project) -> Dataset:
    """
    Retrieves internal dataset required to extract tier-specific average confidence

    Args:
        project: Tamr project object

    Returns:
        Dataset

    Raises:
        RuntimeError: if the required internal dataset does not exist in `project`
    """
    dataset_name = f"{project.unified_dataset().name}_{CONFIDENCE_DATASET_SUFFIX}"
    # check if the dataset can be retrieved
    try:
        dataset = project.client.datasets.by_name(dataset_name)
    except KeyError:
        dataset_not_found_error = (
            "Classification project could be incomplete, or you might be using a version of Tamr "
            "unsupported for this functionality because required dataset could not be found in "
            "the project datasets."
        )
        LOGGER.error(dataset_not_found_error)
        raise RuntimeError(dataset_not_found_error)
    return dataset


def get_tier_confidence(
    project: Project, *, tier: int = -1, allow_dataset_refresh: bool = False
) -> data_type.JsonDict:
    """
    Extracts tier-specific average confidence from a Tamr internal dataset
    `<unified dataset name>_classifications_average_confidences` in a dictionary

    Args:
        project: Tamr project object
        tier: integer specifying the tier to extract the average confidence;
              default value will return the average confidence at all leaf categories
        allow_dataset_refresh: if True, allows running a job to refresh dataset to make it
                               streamable

    Returns:
        dictionary - keys are category paths, joined by '|' if multi-level taxonomy and values are
        average confidence of the corresponding keys

    Raises:
        RuntimeError: if `dataset` is not streamable and `allow_dataset_refresh` is False;
        TypeError: if tier is not of type int;
                   or if the project type is not classification
        ValueError: if tier is less than -1 or equal to 0
    """
    LOGGER.info(
        f"Retrieving average confidence for taxonomy nodes in project {project.name} "
        f"(id={project.resource_id})."
    )
    # check project type is categorization
    try:
        project = project.as_categorization()
    except TypeError:
        not_categorization_error = f"Project {project.name} is not a classification project."
        LOGGER.error(not_categorization_error)
        raise TypeError(not_categorization_error)

    # check necessary dataset can be obtained
    dataset = _get_dataset_with_confidence(project)

    # check tier is valid
    if type(tier) is not int:
        wrong_tier_type_error = f"Tier {tier} is not an integer."
        LOGGER.error(wrong_tier_type_error)
        raise TypeError(wrong_tier_type_error)
    if tier < -1 or tier == 0:
        invalid_tier_value_error = (
            f"Invalid value for tier {tier}. Tier cannot be 0 or less than -1."
        )
        LOGGER.error(invalid_tier_value_error)
        raise ValueError(invalid_tier_value_error)

    # check dataset can be streamed
    if not dataset.status().is_streamable:
        if allow_dataset_refresh:
            LOGGER.info(f"Refreshing dataset {dataset.name} to make streamable.")
            op = dataset.refresh()
            operation.enforce_success(op)
        else:
            cannot_stream_error = (
                f"Dataset {dataset.name} is not streamable. "
                f"Refresh it first, or run with allow_dataset_refresh=True"
            )
            LOGGER.error(cannot_stream_error)
            raise RuntimeError(cannot_stream_error)

    # check dataset contains necessary attributes
    _check_dataset_with_confidence(dataset)

    # check tier does not exceed maximum taxonomy depth
    _check_taxonomy_depth(project, tier=tier)

    # obtain categories at tier
    selected_category_set = _get_categories_at_tier(project, tier=tier)

    # extract average confidence
    tier_confidence_dict = _extract_confidence(dataset=dataset, category_set=selected_category_set)
    return tier_confidence_dict



"""Tasks related to running jobs for Tamr Mastering projects"""
from typing import List
import logging

from tamr_unify_client.mastering.project import MasteringProject
from tamr_unify_client.operation import Operation
from tamr_toolbox.realtime.matching import update_realtime_match_data

from tamr_toolbox.models.project_type import ProjectType
from tamr_toolbox.utils import operation

LOGGER = logging.getLogger(__name__)


def _run_custom(
    project: MasteringProject,
    *,
    run_update_unified_dataset: bool = False,
    run_estimate_pair_counts: bool = False,
    run_generate_pairs: bool = False,
    run_apply_feedback: bool = False,
    run_update_pair_results: bool = False,
    run_update_high_impact_pairs: bool = False,
    run_update_cluster_results: bool = False,
    run_publish_clusters: bool = False,
    run_update_realtime_match: bool = False,
    process_asynchronously: bool = False,
) -> List[Operation]:
    """Executes specified steps of a mastering project.

    Args:
        project: The target mastering project
        run_update_unified_dataset: Whether refresh should be called on the unified dataset
        run_estimate_pair_counts: Whether an estimate pairs job should be run
        run_generate_pairs: Whether refresh should be called on the pairs dataset
        run_apply_feedback: Whether train should be called on the pair matching model
        run_update_pair_results: Whether predict should be called on the pair matching model
        run_update_high_impact_pairs: Whether refresh should be called on the high impact pairs
            dataset
        run_update_cluster_results: Whether refresh should be called on the record clusters dataset
        run_publish_clusters: Whether refresh should be called on the published record clusters
            dataset
        run_update_realtime_match: Whether to update match data with latest published clusters
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run

    Raises:
        TypeError: if the `project` is not a Mastering project
    """
    if ProjectType[project.type] != ProjectType.DEDUP:
        error_msg = f"Cannot use as a mastering project. Project type: {project.type}"
        LOGGER.error(error_msg)
        raise TypeError(error_msg)
    else:
        project = project.as_mastering()

    completed_operations = []
    if run_update_unified_dataset:
        LOGGER.info(
            f"Updating the unified dataset for project {project.name} (id={project.resource_id})."
        )
        op = project.unified_dataset().refresh(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)

    # Run pair-related operations
    completed_operations.extend(
        _run_custom_pair_operations(
            project,
            run_estimate_pair_counts=run_estimate_pair_counts,
            run_generate_pairs=run_generate_pairs,
            run_apply_feedback=run_apply_feedback,
            run_update_pair_results=run_update_pair_results,
            run_update_high_impact_pairs=run_update_high_impact_pairs,
            process_asynchronously=process_asynchronously,
        )
    )

    if run_update_cluster_results:
        LOGGER.info(
            f"Updating cluster prediction results for project {project.name} "
            f"(id={project.resource_id})."
        )
        op = project.record_clusters().refresh(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)
    if run_publish_clusters:
        LOGGER.info(f"Publishing clusters for project {project.name} (id={project.resource_id}).")
        op = project.published_clusters().refresh(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)
    if run_update_realtime_match:
        LOGGER.info(
            f"Updating match database for project {project.name} (id={project.resource_id})."
        )
        op = update_realtime_match_data(project=project, asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)

    return completed_operations


def _run_custom_pair_operations(
    project: MasteringProject,
    *,
    run_estimate_pair_counts: bool = False,
    run_generate_pairs: bool = False,
    run_apply_feedback: bool = False,
    run_update_pair_results: bool = False,
    run_update_high_impact_pairs: bool = False,
    process_asynchronously: bool = False,
) -> List[Operation]:
    """Executes specified steps of a mastering project.

    Args:
        project: The target mastering project
        run_estimate_pair_counts: Whether an estimate pairs job should be run
        run_generate_pairs: Whether refresh should be called on the pairs dataset
        run_apply_feedback: Whether train should be called on the pair matching model
        run_update_pair_results: Whether predict should be called on the pair matching model
        run_update_high_impact_pairs: Whether refresh should be called on the high impact pairs
            dataset
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run

    """

    completed_operations = []
    if run_estimate_pair_counts:
        LOGGER.info(f"Estimate pair counts for project {project.name} (id={project.resource_id}).")
        op = operation.safe_estimate_counts(project)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)
    if run_generate_pairs:
        LOGGER.info(f"Generating pairs for project {project.name} (id={project.resource_id}).")
        op = project.pairs().refresh(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)
    if run_apply_feedback:
        LOGGER.info(
            f"Applying feedback to the pairs model for project {project.name} "
            f"(id={project.resource_id})."
        )
        op = project.pair_matching_model().train(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)
    if run_update_pair_results:
        LOGGER.info(
            f"Updating pair prediction results for project {project.name} "
            f"(id={project.resource_id})."
        )
        op = project.pair_matching_model().predict(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)
    if run_update_high_impact_pairs:
        LOGGER.info(
            f"Refreshing high impact pairs for project {project.name} (id={project.resource_id})."
        )
        op = project.high_impact_pairs().refresh(asynchronous=process_asynchronously)
        if not process_asynchronously:
            operation.enforce_success(op)
        completed_operations.append(op)

    return completed_operations


def run(
    project: MasteringProject,
    *,
    run_estimate_pair_counts: bool = False,
    run_apply_feedback: bool = False,
    run_update_realtime_match: bool = False,
    process_asynchronously: bool = False,
) -> List[Operation]:
    """Run the existing pipeline without training

    Args:
        project: Target mastering project
        run_estimate_pair_counts: Whether an estimate pairs job should be run
        run_apply_feedback: Whether train should be called on the pair matching model
        run_update_realtime_match: Whether to update RealTime match data after publishing clusters
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=True,
        run_estimate_pair_counts=run_estimate_pair_counts,
        run_generate_pairs=True,
        run_apply_feedback=run_apply_feedback,
        run_update_pair_results=True,
        run_update_high_impact_pairs=True,
        run_update_cluster_results=True,
        run_publish_clusters=True,
        run_update_realtime_match=run_update_realtime_match,
        process_asynchronously=process_asynchronously,
    )


def update_unified_dataset(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Updates the unified dataset for a mastering project

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=True,
        run_estimate_pair_counts=False,
        run_generate_pairs=False,
        run_apply_feedback=False,
        run_update_pair_results=False,
        run_update_cluster_results=False,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def estimate_pair_counts(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Estimates the number of pairs for a mastering project

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=True,
        run_generate_pairs=False,
        run_apply_feedback=False,
        run_update_pair_results=False,
        run_update_cluster_results=False,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def generate_pairs(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Generates the pairs for a mastering project

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=False,
        run_generate_pairs=True,
        run_apply_feedback=False,
        run_update_pair_results=False,
        run_update_cluster_results=False,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def apply_feedback(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """
    Applies feedback to update the model for a mastering project

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=False,
        run_generate_pairs=False,
        run_apply_feedback=True,
        run_update_pair_results=False,
        run_update_cluster_results=False,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def update_pair_predictions(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """
    Updates pair predictions only.

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=False,
        run_generate_pairs=False,
        run_apply_feedback=False,
        run_update_pair_results=True,
        run_update_cluster_results=False,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def update_clusters(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """
    Re-runs clustering only.

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=False,
        run_generate_pairs=False,
        run_apply_feedback=False,
        run_update_pair_results=False,
        run_update_cluster_results=True,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def apply_feedback_and_update_results(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Trains the model, predicts the pair labels, and updates the draft clusters of
    a mastering project

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=False,
        run_generate_pairs=False,
        run_apply_feedback=True,
        run_update_pair_results=True,
        run_update_high_impact_pairs=True,
        run_update_cluster_results=True,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def update_results_only(
    project: MasteringProject, *, process_asynchronously: bool = False
) -> List[Operation]:
    """Predicts the pair labels based on the existing pair model and updates the draft clusters
    of a mastering project

    Args:
        project: Target mastering project
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=False,
        run_generate_pairs=False,
        run_apply_feedback=False,
        run_update_pair_results=True,
        run_update_high_impact_pairs=True,
        run_update_cluster_results=True,
        run_publish_clusters=False,
        process_asynchronously=process_asynchronously,
    )


def publish_clusters(
    project: MasteringProject,
    *,
    run_update_realtime_match: bool = False,
    process_asynchronously: bool = False,
) -> List[Operation]:
    """Publishes the clusters of a mastering project

    Args:
        project: Target mastering project
        run_update_realtime_match: whether to update RealTime match data after publishing clusters
        process_asynchronously: Whether or not to wait for the job to finish before returning
            - must be set to True for concurrent workflow

    Returns:
        The operations that were run
    """
    return _run_custom(
        project,
        run_update_unified_dataset=False,
        run_estimate_pair_counts=False,
        run_generate_pairs=False,
        run_apply_feedback=False,
        run_update_pair_results=False,
        run_update_cluster_results=False,
        run_publish_clusters=True,
        run_update_realtime_match=run_update_realtime_match,
        process_asynchronously=process_asynchronously,
    )



"""Tasks related to moving data in or out of Tamr using pandas.DataFrames"""
import os
import logging

from typing import Optional, List, Callable, Any, Iterable, Tuple
from collections import OrderedDict, defaultdict
from functools import partial

from tamr_unify_client.dataset.resource import Dataset

from tamr_toolbox.data_io import common
from tamr_toolbox.models.validation_check import ValidationCheck
from tamr_toolbox.utils import operation


LOGGER = logging.getLogger(__name__)

# Building our documentation requires access to all dependencies, including optional ones
# This environments variable is set automatically when `invoke docs` is used
BUILDING_DOCS = os.environ.get("TAMR_TOOLBOX_DOCS") == "1"
if BUILDING_DOCS:
    # Import relevant optional dependencies
    import pandas


def _flatten_list_series(
    series: "pandas.Series",
    *,
    delimiter: str = "|",
    force: bool = False,
    columns: Optional[List] = None,
) -> "pandas.Series":
    """
    Flattens list types to strings.  See ``_flatten_list`` for details.
    Operates on a pandas Series, and is intended to be applied using `apply` to a pandas DataFrame.

    Args:
        series: pandas Series of values to flatten
        delimiter: string to use as delimiter for concatenating lists
        force: if True, will force non-string inner types to string
        columns: optional, list of columns to flatten

    Returns: Series of flattened values
    """
    func = partial(common._flatten_list, delimiter=delimiter, force=force)
    # if we haven't specified which columns, or this column is in the list, apply the function
    if columns is None or series.name in columns:
        return series.apply(func)
    # otherwise do nothing
    else:
        return series


def flatten(
    df: "pandas.DataFrame",
    *,
    delimiter: str = "|",
    columns: Optional[List[str]] = None,
    force: bool = False,
) -> "pandas.DataFrame":
    """
    Converts DataFrame columns with list type to strings
    and returns a copy of the DataFrame with this change.
    Tamr often produces datasets with columns of list type,
    which often are more easily consumed as single-valued columns.

    Args:
        df: DataFrame from Tamr dataset
        delimiter: string to use as delimiter for concatenating lists
        columns: optional, list of columns to flatten
        force: if True, will force non-string inner types to string

    Returns:
        flattened DataFrame
    """
    # apply a series-wise function that can operate on specific columns of the dataframe
    series_func = partial(_flatten_list_series, delimiter=delimiter, force=force, columns=columns)
    return df.apply(series_func)


def from_dataset(
    dataset: Dataset,
    *,
    columns: Optional[List[str]] = None,
    flatten_delimiter: Optional[str] = None,
    flatten_columns: Optional[List[str]] = None,
    force_flatten: bool = False,
    nrows: Optional[int] = None,
    allow_dataset_refresh: bool = False,
) -> "pandas.DataFrame":
    """
    Creates a DataFrame from a Tamr Dataset

    Args:
        dataset: Tamr Dataset object
        columns: optional, ordered list of columns to keep
        flatten_delimiter: if set, flatten list types to strings by concatenating with this
            delimiter
        flatten_columns: optional, list of columns to flatten
        force_flatten:  if False, arrays with inner types other than string will not be flattened.
            if True, will force all inner types to strings when flattening values.
            if True, `flatten_delimiter` must be specified.
        nrows: number of rows to read. default None will read all rows
        allow_dataset_refresh: if True, allows running a job to refresh dataset to make streamable

    Returns:
        DataFrame

    Raises:
        ValueError: if `columns` or `flatten_columns` contain columns that are not present in
            `dataset`
    """
    # Validate arguments
    if force_flatten and flatten_delimiter is None:
        raise ValueError("`force_flatten` is True but no `flatten_delimiter` is specified.")

    # This function requires pandas, an optional dependency
    import pandas

    LOGGER.info(
        f"Streaming records to DataFrame for dataset {dataset.name} (id={dataset.resource_id})."
    )
    dataset_attrs = [attr for attr in dataset.attributes]
    attr_names = [attr.name for attr in dataset_attrs]
    # check that specified columns exist
    if columns is not None:
        common._check_columns_subset(
            input_list=columns, reference_list=attr_names, raise_error=True
        )
    # checks on columns to flatten
    if flatten_delimiter is not None:
        if flatten_columns is None:
            flatten_columns = list(attr_names)
        else:
            # check that specified columns exist
            common._check_columns_subset(
                input_list=flatten_columns, reference_list=attr_names, raise_error=True
            )
        # check types of flatten_columns
        for attr in dataset_attrs:
            if attr.name not in flatten_columns:
                continue
            attr_type = attr.spec().to_dict()["type"]
            if attr_type["baseType"] == "ARRAY" and attr_type["innerType"]["baseType"] != "STRING":
                if force_flatten:
                    LOGGER.info(
                        f"Will force attribute to string: {attr.name}, with type: {attr_type}"
                    )
                else:
                    LOGGER.warning(
                        f"Will not flatten attribute: {attr.name}, with type: {attr_type}"
                    )
                    flatten_columns.remove(attr.name)

    if not dataset.status().is_streamable:
        if allow_dataset_refresh:
            LOGGER.info(f"Refreshing dataset {dataset.name} to make streamable.")
            op = dataset.refresh()
            operation.enforce_success(op)
        else:
            message = (
                f"Dataset {dataset.name} is not streamable. Refresh it first, or run"
                f" with allow_dataset_refresh=True"
            )
            LOGGER.error(message)
            raise RuntimeError(message)

    # if flattening, set the function to apply to records as _flatten_list
    # otherwise set as _identity
    func = None
    if flatten_delimiter is not None:
        func = partial(common._flatten_list, delimiter=flatten_delimiter, force=force_flatten)
    df = pandas.DataFrame.from_records(
        common._yield_records(
            dataset, func=func, columns=columns, flatten_columns=flatten_columns
        ),
        columns=columns,
        nrows=nrows,
    )
    return df


def _count_null(s: "pandas.Series") -> int:
    """
    Returns the number of null values in a pandas Series

    Args:
        s: pandas Series
    Returns:
        int
    """
    return s.isna().sum()


def profile(df: "pandas.DataFrame") -> "pandas.DataFrame":
    """
    Computes profile statistics from an input DataFrame,
    and returns statistics in another DataFrame.
    Intended to be used for validation checks on a DataFrame before
    upserting records to a Tamr Dataset

    Args:
        df: DataFrame

    Returns:
        DataFrame with profile statistics
    """
    # compute per-attribute metrics
    # attr_metrics: aggregations that return a Series indexed by the attribute names
    # attr_metric_names: aggregations that return a Series indexed by the attribute names
    attr_agg_functions = OrderedDict()

    # number of unique values
    attr_agg_functions["DistinctValueCount"] = "nunique"

    # number of null values
    attr_agg_functions["EmptyValueCount"] = _count_null

    # aggregate to compute per-attribute metrics
    df_profile = df.agg(list(attr_agg_functions.values()), axis=0).T
    # rename columns
    df_profile.columns = attr_agg_functions.keys()

    # add summary level metrics
    df_profile["RecordCount"] = df.shape[0]

    return df_profile


def _check_present_columns(
    df: "pandas.DataFrame", *, require_present_columns: Optional[List[str]] = None
) -> ValidationCheck:
    """
    Checks that a specified list of columns in a DataFrame are all present

    Args:
        df: DataFrame
        require_present_columns: list of columns that are checked to be present

    Returns:
        ValidationCheck object, with bool for whether check passed and dict of failing columns
    """
    failed_checks_dict = defaultdict(list)

    if require_present_columns is None:
        require_present_columns = []

    # check if columns are present
    for col in require_present_columns:
        if col not in df.columns:
            LOGGER.warning(f"column {col} is not present")
            failed_checks_dict["failed_present_columns"].append(col)

    passed = len(failed_checks_dict) == 0
    return ValidationCheck(passed, failed_checks_dict)


def _check_unique_columns(
    df_profile: "pandas.DataFrame", *, require_unique_columns: Optional[List[str]] = None
) -> ValidationCheck:
    """
    Checks that a specified list of columns in a DataFrame have all unique values

    Args:
        df_profile: DataFrame containing profile information
        require_unique_columns: list of columns that are checked to have all unique values,
            e.g. a primary key column

    Returns:
        ValidationCheck object, with bool for whether check passed and dict of failing columns
    """
    failed_checks_dict = defaultdict(list)

    if require_unique_columns is None:
        require_unique_columns = []

    # check for unique columns
    for col in require_unique_columns:
        num_unique = df_profile.loc[col, "DistinctValueCount"]
        num_records = df_profile.loc[col, "RecordCount"]
        if not num_unique == num_records:
            LOGGER.warning(
                f"column {col} has only {num_unique} unique values out of {num_records} records"
            )
            failed_checks_dict["failed_unique_columns"].append(col)

    passed = len(failed_checks_dict) == 0
    return ValidationCheck(passed, failed_checks_dict)


def _check_nonnull_columns(
    df_profile: "pandas.DataFrame", *, require_nonnull_columns: Optional[List[str]] = None
) -> ValidationCheck:
    """
    Checks that a specified list of columns in a DataFrame have all non-null values

    Args:
        df_profile: DataFrame containing profile information
        require_nonnull_columns: list of columns that are checked to have all non-null values

    Returns:
        ValidationCheck object, with bool for whether check passed and dict of failing columns
    """
    failed_checks_dict = defaultdict(list)

    if require_nonnull_columns is None:
        require_nonnull_columns = []

    # check for nonnull columns
    for col in require_nonnull_columns:
        num_null = df_profile.loc[col, "EmptyValueCount"]
        if num_null > 0:
            LOGGER.warning(f"column {col} has {num_null} null values")
            failed_checks_dict["failed_nonnull_columns"].append(col)

    passed = len(failed_checks_dict) == 0
    return ValidationCheck(passed, failed_checks_dict)


def _check_custom(
    df: "pandas.DataFrame",
    columns_to_check: Optional[List[str]],
    check_function: Callable[[Any], bool],
) -> ValidationCheck:
    """
    Args:
        df: Dataframe
        columns_to_check: columns on which check_function will be applied
        check_function: function applied on columns_to_check

    Returns:
        ValidationCheck object, with bool for whether all checks passed and dict of failing columns
    """
    failed_checks_dict = defaultdict(list)

    df1 = df[columns_to_check].applymap(check_function)
    for col in columns_to_check:
        if not df1[col].all():
            LOGGER.warning(f"column {col} failed custom check {check_function.__name__}")
            failed_checks_dict[f"failed custom check {check_function.__name__}"].append(col)

    passed = len(failed_checks_dict) == 0
    return ValidationCheck(passed, failed_checks_dict)


def validate(
    df: "pandas.DataFrame",
    *,
    raise_error: bool = True,
    require_present_columns: Optional[List[str]] = None,
    require_unique_columns: Optional[List[str]] = None,
    require_nonnull_columns: Optional[List[str]] = None,
    custom_checks: Iterable[Tuple[Callable[[Any], bool], List[str]]] = tuple(),
) -> ValidationCheck:
    """
    Performs validation checks on a DataFrame.
    Returns a dict of columns that fail each check, and optionally returns an error.
    Intended to be used on a DataFrame prior to upserting records into a Tamr dataset.

    Args:
        df: DataFrame
        raise_error: if True, will raise a ValueError on failing checks.  if False,
            will print Warning and return a dict
        require_present_columns: list of columns that are checked to be present
        require_unique_columns: list of columns that are checked to have all unique values,
            e.g. a primary key column
        require_nonnull_columns: list of columns that are checked to have all non-null values
        custom_checks: collection of tuples each containing a custom function and list of columns,
            on which to apply it


    Returns:
        ValidationCheck object, with bool for whether all checks passed and dict of failing columns

    Raises:
        ValueError: if `raise_error` is set True, and any checks fail
    """
    failed_checks_dict = defaultdict(list)

    if require_present_columns is None:
        require_present_columns = []
    if require_unique_columns is None:
        require_unique_columns = []
    if require_nonnull_columns is None:
        require_nonnull_columns = []

    # compute profile stats
    df_profile = profile(df)

    # check for present columns
    failed_checks_dict.update(
        _check_present_columns(df, require_present_columns=require_present_columns).details
    )

    # check for unique columns
    failed_checks_dict.update(
        _check_unique_columns(df_profile, require_unique_columns=require_unique_columns).details
    )

    # check for nonnull columns
    failed_checks_dict.update(
        _check_nonnull_columns(df_profile, require_nonnull_columns=require_nonnull_columns).details
    )

    for custom_check in custom_checks:
        failed_checks_dict.update(
            _check_custom(
                df, check_function=custom_check[0], columns_to_check=custom_check[1]
            ).details
        )

    failed_checks_dict = dict(failed_checks_dict)
    passed = len(failed_checks_dict) == 0
    if not passed and raise_error:
        raise (
            ValueError(
                f"DataFrame validation failed.  failed columns for each check: "
                f"{failed_checks_dict}"
            )
        )
    return ValidationCheck(passed, failed_checks_dict)



"""Tasks related to Tamr projects"""
from . import attribute_configuration
from . import schema
from . import transformations
from .movement import export_artifacts, import_artifacts

__all__ = [
    "attribute_configuration",
    "schema",
    "transformations",
    "export_artifacts",
    "import_artifacts",
]



"""Tasks related to transformations with Tamr projects"""
from typing import List
from dataclasses import dataclass, field
from requests import Response
import logging

from tamr_unify_client import response, Client
from tamr_unify_client.dataset.resource import Dataset
from tamr_unify_client.project.resource import Project

from tamr_toolbox.models.data_type import JsonDict
from tamr_toolbox.utils.version import requires_tamr_version

LOGGER = logging.getLogger(__name__)

TX_RELEASE_VERSION = "2020.009.0"


@dataclass()
class InputTransformation:
    """A transformation scoped to input datasets

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        transformation: The text of a transformations script
        datasets: The list of input datasets that the script should be applied to
    """

    transformation: str
    datasets: List[Dataset] = field(default_factory=list)


@dataclass()
class TransformationGroup:
    """A group of input transformations and unified transformations

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        input_scope: A list of transformation to apply to input datasets
        unified_scope: A list of transformation scripts to apply to the unified dataset
    """

    input_scope: List[InputTransformation] = field(default_factory=list)
    unified_scope: List[str] = field(default_factory=list)


def _input_transformation_from_json(tamr_client: Client, data: JsonDict) -> InputTransformation:
    """Make input transformation from JSON data (deserialize)

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        tamr_client: A Tamr client
        data: Input scoped transformation JSON data from Tamr server

    Returns:
        A single input transformation
    """
    # "datasets" is a dictionary with 3 keys: name, datasetId and relativeDatasetId
    # The datasetId is formatted like: "datasetId": "unify://unified-data/v1/datasets/3"
    # If we split that string on / , the last value will be the dataset resource id
    dataset_resource_ids = [d["datasetId"].split("/")[-1] for d in data["datasets"]]
    datasets = [tamr_client.datasets.by_resource_id(d_id) for d_id in dataset_resource_ids]
    # "transformation" is the text of the transformation script
    return InputTransformation(transformation=data["transformation"], datasets=datasets)


def _from_json(tamr_client: Client, data: JsonDict) -> TransformationGroup:
    """Make transformations from JSON data (deserialize)

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        tamr_client: A Tamr client
        data: Transformation JSON data from Tamr server

    Returns:
        All input transformations and unified transformations
    """
    return TransformationGroup(
        # "unified" is a list of transformations for the unified dataset
        unified_scope=data["unified"],
        # "parameterized" is a list of dictionaries representing input transformations
        input_scope=[
            _input_transformation_from_json(tamr_client, tx) for tx in data["parameterized"]
        ],
    )


def _input_transformation_to_json(tx: InputTransformation) -> JsonDict:
    """Convert input transformations to JSON data (serialize)

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        tx: Input transformation to convert

    Returns:
        JSON representation of a single input transformation
    """
    # datasetId omitted, only one of "datasetId" or "relativeDatasetId" is required
    dataset_json = [
        {"name": d.name, "relativeDatasetId": f"datasets/{d.resource_id}"} for d in tx.datasets
    ]

    return {"datasets": dataset_json, "transformation": tx.transformation}


def _to_json(tx: TransformationGroup) -> JsonDict:
    """Convert transformations to JSON data (serialize)

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        tx: Transformations to convert

    Returns:
        JSON representation of all input transformations and unified transformations

    """
    return {
        "parameterized": [_input_transformation_to_json(t) for t in tx.input_scope],
        "unified": tx.unified_scope,
    }


@requires_tamr_version(min_version=TX_RELEASE_VERSION)
def get_all(project: Project) -> TransformationGroup:
    """Get the transformations of a Project

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        project: Project containing transformations

    Returns:
        All input transformations and unified transformations of a project
    """

    LOGGER.info(f"Getting transformations from project {project}")
    r = project.client.get(f"{project.api_path}/transformations")
    response.successful(r)
    return _from_json(project.client, r.json())


@requires_tamr_version(min_version=TX_RELEASE_VERSION)
def set_all(project: Project, tx: TransformationGroup, *, allow_overwrite=True) -> Response:
    """Set the transformations of a Project

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        project: Project to place transformations within
        tx: Transformations to put into project
        allow_overwrite: Whether existing transformations can be overwritten

    Returns:
        Response object created when transformations of a project are replaced

    Raises:
        RuntimeError: if `allow_overwrite` is set to False but transformations already exists
            in `project`
        ValueError: if provided `tx` are invalid
    """

    if not allow_overwrite:
        # Fail if any transformations already exist
        current_tx = get_all(project)
        if (len(current_tx.unified_scope) + len(current_tx.input_scope)) != 0:
            error_message = (
                f"Cannot set transformations in project {project} because "
                f"transformations already exist in that project and allow_overwrite "
                f"is set to False.)"
            )
            LOGGER.error(error_message)
            raise RuntimeError(error_message)

    LOGGER.info(f"Replaces transformations in project {project}")
    body = _to_json(tx)
    r = project.client.put(f"{project.api_path}/transformations", json=body)

    # Raise a specific exception for invalid transformation input
    if r.status_code == 400:
        error_message = (
            f"Failed to set transformations in project {project} due to invalid "
            f"transformations: {r.json()['message']}"
        )
        LOGGER.error(error_message)
        raise ValueError(error_message)

    # Other issue with set are caught here
    return response.successful(r)


def get_all_unified(project: Project) -> List[str]:
    """Get the unified transformations of a Project

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        project: Project containing transformations

    Returns:
        All unified transformations of a project
    """
    return get_all(project).unified_scope


def set_all_unified(project: Project, tx: List[str], *, allow_overwrite=True) -> Response:
    """Set the unified transformations of a Project.
    Any input transformations will not be altered

    Version:
        Requires Tamr 2020.009.0 or later

    Args:
        project: Project to place transformations within
        tx: Unified transformations to put into project
        allow_overwrite: Whether existing unified transformations can be overwritten

    Returns:
        Response object created when transformations of a project are replaced

    Raises:
        RuntimeError: if `allow_overwrite` is set to False but transformations already exists
            in `project`
    """
    current_tx = get_all(project)

    if not allow_overwrite:
        # Fail if any unified transformations already exist
        if len(current_tx.unified_scope) != 0:
            error_message = (
                f"Cannot set unified transformations in project {project} because "
                f"unified transformations already exist in that project and "
                f"allow_overwrite is set to False.)"
            )
            LOGGER.error(error_message)
            raise RuntimeError(error_message)

    new_tx = TransformationGroup(unified_scope=tx, input_scope=current_tx.input_scope)

    # Here allow_overwrite is true, as input transformations are allowed to be present
    return set_all(project, new_tx, allow_overwrite=True)
